@@ -1,6 +1,6 @@
 /* arping/src/findif_other.c
  *
- *  Copyright (C) 2000-2009 Thomas Habets <thomas@habets.pp.se>
+ *  Copyright (C) 2000-2009 Thomas Habets <thomas@habets.se>
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public
@@ -29,10 +29,13 @@
  *
  */
 const char *
-arping_lookupdev(const char *ifname,
-                 uint32_t srcip,
-                 uint32_t dstip,
-                 char *ebuf)
+arping_lookupdev(uint32_t srcip, uint32_t dstip, char *ebuf)
 {
-	return arping_lookupdev_default(ifname,srcip,dstip,ebuf);
+	return NULL;
 }
+/* ---- Emacs Variables ----
+ * Local Variables:
+ * c-basic-offset: 8
+ * indent-tabs-mode: nil
+ * End:
+ */
